/*
 * backend.c - cvs administration "library"
 *
 * Copyright (c) 2001, Vincent Labrecque <limitln@@cooptel.qc.ca>
 * All rights reserved.
 * 
 *      Redistribution and use in source and binary forms,
 *      with or without modification, are permitted provided
 *      that the following conditions are met:
 * 
 *      Redistributions of source code must retain the
 *      above copyright notice, this list of conditions
 *      and the following disclaimer. 
 *      Redistributions in binary form must reproduce
 *      the above copyright notice, this list of
 *      conditions and the following disclaimer in the
 *      documentation and/or other materials provided
 *      with the distribution. 
 *      Neither the name of the <ORGANIZATION> nor the
 *      names of its contributors may be used to
 *      endorse or promote products derived from this
 *      software without specific prior written
 *      permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "backend.h"
#include "util.h"
#include "xmalloc.h"

/* Some data structures */

struct ca_hfile {
	struct ca_hfile *next;
	struct ca_db *db;
	char *fpath;
	int (*r)();
	int (*w)();
};

struct ca_user {
	struct ca_user *next;
	char *name;
	char *pass;
	char *sysuser;
	char *email;
};

struct ca_db {
	bool changed;
	bool usecvs;
	char *container;
	char *repository;
	struct ca_hfile *hfiles;
	struct ca_user *users;
	int nusers;
};

/* Private functions */
static int		 handle_pass_line __P((int, char *));

static int		 passwd_read __P((struct ca_hfile *));
static int		 passwd_write __P((struct ca_hfile *));
static int		 users_read __P((struct ca_hfile *));
static int		 users_write __P((struct ca_hfile *));
static int		 etc_cvs_passwd_read __P((struct ca_hfile *));
static int		 etc_cvs_passwd_write __P((struct ca_hfile *));

static struct ca_db	*get_db __P((int));
static int		 get_id __P((struct ca_db *));

static struct ca_db	*ca_db_new __P((char *repos, bool));
static void		 ca_db_free __P((struct ca_db *));
static int		 ca_db_try_hfile __P((struct ca_db *, char *, 
					      int, int (*)(), int (*)()));
static int		 ca_db_del_hfile __P((struct ca_db *, struct ca_hfile *));
static struct ca_user	*ca_db_find_user __P((struct ca_db *db, const char *));
static int		 ca_db_add_user __P((struct ca_db *db, struct ca_user *));
static int		 ca_db_del_user __P((struct ca_db *db, struct ca_user *));

static struct ca_hfile	*ca_hfile_new __P((struct ca_db *, char *,
					   int, int (*)(), int (*)()));
static int		 ca_hfile_read __P((struct ca_hfile *));
static int		 ca_hfile_write __P((struct ca_hfile *));
static void		 ca_hfile_free __P((struct ca_hfile *));

static struct ca_user	*ca_user_new __P((char *, char *, char *, char *));
static void		 ca_user_free __P((struct ca_user *));

static int		 is_local_repository __P((const char *));

/*
 * File parsing
 */

/*
 * handle_pass_line
 *
 * Handle a single password file line
 *
 * This is in a separate function because /etc/cvs.passwd and CVSROOT/passwd
 * have the same format (user:pass:sysuser)
 */
static int
handle_pass_line(int db_id, char *line)
{
	char *stuff[3], *p, *pp;

	/* OK, This is not really cute :). */
	p = index(line, ':'); if(!p)  return -1;
	stuff[0] = line; *p++ = '\0'; pp = p;
	p = index(pp, ':');
	if (p != NULL)
		*p++ = '\0';

	stuff[1] = pp;
	stuff[2] = p;

	if (cvsadmin_set_user_info(db_id, stuff[0],
				   stuff[1], stuff[2], NULL) < 0) {
		fprintf(stderr, "INTERNAL ERROR: %s: _set_user_info problem\n", __FUNCTION__);
		return -1;
	}

	return 0;
}

/***************************** CVSROOT/passwd ***********/

/*
 * passwd_read
 *
 * Read <hf> as a password file.  We insert users in <hf>'s database, etc.
 */
static int
passwd_read(struct ca_hfile *hf)
{
	FILE *in;
	char buf[BUFSIZ], *line;
	int ret, db, lineno;

	in = fopen(hf->fpath, "r");
	if (in == NULL) {
		perror(hf->fpath);
		return -1;
	}

	ret = 0;
	lineno = 0;
	db = get_id(hf->db);

	while (fgets(buf, sizeof(buf), in) != NULL) {
		lineno++;
		line = clean_string(buf);
		if (line == NULL || *line == '#')
			continue;
	
		if (handle_pass_line(db, line) < 0) {
			fprintf(stderr, "Invalid line %d in password file\n",
				lineno);
			ret = -1;
			break;
		}
	}	
	if (fclose(in) < 0)
		perror("fclose");

	return ret;
}

/*
 * passwd_write
 *
 * Dump out all stuff belonging in a password file in <hf>'s db in <hf>
 */
static int
passwd_write(struct ca_hfile *hf)
{
	FILE *out;
	struct ca_user *p;

	out = fopen(hf->fpath, "w");
	if (out == NULL) {
		perror(hf->fpath);
		return -1;
	}
	
	for (p = hf->db->users; p != NULL; p = p->next) {
		fprintf(out, "%s:%s:%s\n",
			p->name, p->pass?p->pass:"", p->sysuser?p->sysuser:"");
	}
	
	if (fclose(out) < 0)
		perror("fclose");

	return 0;
}


/***************************** CVSROOT/users ***********/

/*
 * users_read
 *
 * Read <hf> as a /users file.
 */
static int
users_read(struct ca_hfile *hf)
{
	FILE *in;
	char buf[BUFSIZ], *line, *p;
	int lineno, ret, db;

	in = fopen(hf->fpath, "r");
	if (in == NULL) {
		perror(hf->fpath);
		return -1;
	}

	ret = 0;
	lineno = 0;
	db = get_id(hf->db);

	while (fgets(buf, sizeof(buf), in) != NULL) {
		lineno++;
		line = clean_string(buf);
		if (line == NULL || *line == '#')
			continue;
		
		p = index(line, ':');
		if (p == NULL) {
			fprintf(stderr, "Invalid line %d in users file\n",
				lineno);
			ret = -1;
			break;
		}
		*p++ = '\0';

		/* if user doesn't exist, don't set the email address. */
		if (cvsadmin_get_user_info(db, line, NULL, NULL, NULL) < 0)
			continue;
		cvsadmin_set_user_info(db, line, NULL, NULL, p);
	}

	if (fclose(in) < 0)
		perror("fclose");

	return ret;
}

/*
 * users_write
 *
 * Dump all information belonging in a /users file in <hf>
 */
static int
users_write(struct ca_hfile *hf)
{
	FILE *out;
	struct ca_user *p;

	out = fopen(hf->fpath, "w");
	if (out == NULL) {
		perror(hf->fpath);
		return -1;
	}

	for (p = hf->db->users; p != NULL; p = p->next){
		if (p->email)
			fprintf(out, "%s:%s\n",	p->name, p->email);
	}

	if (fclose(out) < 0)
		perror("fclose");

	return 0;
}

/***************************** /etc/cvs.passwd ***********/

/*
 * etc_cvs_passwd_read
 *
 * Read <hf> as a /etc/cvs.passwd file.
 */
static int
etc_cvs_passwd_read(struct ca_hfile *hf)
{
	FILE *in;
	char buf[BUFSIZ], *repos, *line;
	int db, lineno, ret;

	/* pretend we didn't find it, it's not related to this host, so just
	 * return */
	if (hf->db->usecvs) {
		if (hf->db->repository[0] != '/' &&
		    strncmp(hf->db->repository, ":local:", 7) != 0)
			return -1;
	}

	repos = extract_path(hf->db->repository);

	in = fopen(hf->fpath, "r");
	if (in == NULL) {
		perror(hf->fpath);
		return -1;
	}

	db = get_id(hf->db);

	ret = lineno = 0;

	while (fgets(buf, sizeof(buf), in) != NULL) {
		lineno++;
		line = clean_string(buf);
		if (line == NULL || *line == '#') continue;

		/* OK, we found our part of the file! */
		if (!strcmp(line, repos)) {
			while (fgets(buf, sizeof(buf), in) != NULL &&
			       buf[0] != '\n'
			       /* && buf[0] != '/') not sure */ ) {
				line = clean_string(buf);
				if (line == NULL || *line == '#')
					continue;
				if (handle_pass_line(db, line) < 0) {
					fprintf(stderr, "Invalid line %d in /etc/cvs.passwd\n",
						lineno);
					ret = -1;
					break;
				}
			}
			break;
		}
	}

	if (fclose(in) < 0) perror("closing");

	return ret;
}

/*
 * etc_cvs_passwd_write
 *
 * Write out the /etc/cvs.passwd file.
 */
static int
etc_cvs_passwd_write(struct ca_hfile *hf)
{
	/*
	 * Create a temp file. Dump the data until we hit our part
	 * of the password file, or EOF.  Then dump our users info,
	 * then dump the rest of the file.  Then, of course, rename
	 * the temp file to /etc/cvs.passwd.
	 */
	FILE *in, *tmp;
	char buf[BUFSIZ], tmpfname[]="/tmp/.ca_tmXXXXXX", *line;
	int tmpfd, done;
	char *repos;
	struct ca_user *u;

	repos = extract_path(hf->db->repository);

	in = fopen(hf->fpath, "r");
	if (in == NULL) {
		perror(hf->fpath);
		return -1;
	}
	tmpfd = mkstemp(tmpfname);
	if (tmpfd < 0) {
		perror("mkstemp");
		if (fclose(in) < 0)
			perror("fclose");
		return -1;
	}
	tmp = fdopen(tmpfd, "w");
	if (tmp == NULL) {
		perror("fdopen (big trouble in hyperspace!)");
		if (fclose(in) < 0)
			perror("fclose");
		return -1;
	}

	done = 0;

	/* 
	 * This trashes redundant space characters from the original file but..
	 * who cares?
	 */
	while (fgets(buf, sizeof(buf), in) != NULL) {
		line = clean_string(buf);
		if (line == NULL)
			line = "";
		if (!done && !strcmp(line, repos)) {
			fprintf(tmp, "%s\n", line);

			for(u = hf->db->users; u!= NULL; u= u->next){
				fprintf(tmp, "%s:%s:%s\n",
					u->name,
					u->pass ? u->pass : "",
					u->sysuser ? u->sysuser : "");
			}
			/* Create an empty space for cleanness */
			fprintf(tmp, "\n");

			done++;
		} else
			fprintf(tmp, "%s\n", line);
	}
	if (fclose(in) < 0)
		perror("fclose");
	if (fclose(tmp) < 0)
		perror("fclose");

	/* there is a window where the temp file could be changed, here ...*/

	if (my_rename(tmpfname, "/etc/cvs.passwd") < 0) {
		fprintf(stderr, "Error renaming our temp file to/etc/cvs.passwd\n"
			"so changes will _not_ be saved (we'll remove the temp file\n"
			"of course. Error was: ");
		perror(NULL);
		if (unlink(tmpfname) < 0)
			fprintf(stderr, "I'm afraid we couldn't do that. Could you\n"
				"please remove %s ASAP? (it contains passwords!)\n",
				tmpfname);
		return -1;
	}

	return 0;
}


/*
 * Internal data structures code
 */


/*
 * ca_db_new
 *
 * Create a new repository db.
 */
static struct ca_db *
ca_db_new(char *repos, bool use_cvs)
{
	struct ca_db *db;

	db = malloc(sizeof(struct ca_db));
	if (db == NULL) {
		perror("malloc");
		return(NULL);
	}

	db->repository = strdup(repos);

	if (use_cvs) {
		db->container = get_temp_dir();	 /* strdup()ed, can free() */
		if (call_cvs4(db->container, db->repository, "checkout",
			      "CVSROOT") < 0) {
			free(db->container);
			free(db->repository);
			return(NULL);
		}
	}
	else
		db->container = strdup(extract_path(db->repository));

	db->usecvs = use_cvs;
	db->changed = 0;
	db->users = NULL;
	db->hfiles = NULL;

	return db;
}

/*
 * ca_db_free
 *
 * Free internal resources allocated to <db>
 */
static void
ca_db_free(struct ca_db *db)
{
	free(db->container);
	free(db->repository);
	free(db);
}

/*
 * ca_db_try_hfile
 *
 * Try to create a hfile in <db> from <path> with [rw] methods <r> and <w>
 *
 * If the <create> flag is on, we force creation of this file.  This should
 * be used for the fallback hfile.
 *
 * Returns 1 on success and 0 on failure.
 */
static int
ca_db_try_hfile(struct ca_db *db, char *path, int create,
		int (*r)(), int (*w)())
{
	char buf[FILENAME_MAX];
	int remain;
	struct ca_hfile *hf;


	buf[0] = '\0';		/* We will strcat! */
	remain = sizeof(buf)-1;

	if (path[0] != '/') {
		strncpy(buf,db->container,remain);
		remain -= strlen(db->container);
		strncat(buf, "/CVSROOT/", remain);
		remain -= 9;
	}
	strncat(buf, path, remain);

	if (is_file(buf) || errno == ENOENT) {
		hf = ca_hfile_new(db, buf, create, r, w);
		if (hf == NULL) {
			return 0;
		}
		if (ca_hfile_read(hf) < 0) {
			fprintf(stderr, "Couldn't hfile_read\n");
			ca_hfile_free(hf);
			return 0;
		}

		return 1;
	}

	return 0;
}

/* 
 * ca_db_del_hfile
 *
 * The db is not necessary, hf->db would do it, but well, it wouldn't
 * fit as nicely with the rest of the API
 */
static int
ca_db_del_hfile(struct ca_db *db, struct ca_hfile *hf)
{
	struct ca_hfile *p, *prev;

	for (p = db->hfiles, prev = NULL; p != NULL; prev = p, p = p->next)
		if (p == hf) {
			if (prev != NULL)
				prev->next = p->next;
			else
				db->hfiles = p->next;
			return 0;
		}
	return -1;
}

/*
 * ca_db_find_user
 *
 * Find given user in the repository.
 */
static struct ca_user *
ca_db_find_user(struct ca_db *db, const char *nam)
{
	struct ca_user *p;

	for (p = db->users; p!=NULL; p=p->next){
		if (!strcmp(p->name, nam))
			break;
	}
	return p;
}

static int
ca_db_add_user(struct ca_db *db, struct ca_user *user)
{
	struct ca_user *end;

	/* Simple linked list insertion */

	for (end = db->users; end && end->next; end=end->next)
		;		/* Nothing */
	if (end == NULL)
		db->users = user;
	else
		end->next = user;

	db->nusers++; 
	db->changed++;

	return 0;
}

static int
ca_db_del_user(struct ca_db *db, struct ca_user *user)
{
	struct ca_user *p, *prev;

	/* Simply linked-list deletion */
	for (p = db->users, prev = NULL; p != NULL; prev = p, p = p->next)
		if (p == user) {
			if (prev != NULL)
				prev->next = p->next;
			else
				db->users = p->next;
			db->nusers--;
			db->changed++;

			return 0;
		}
	return -1;	/* user not found */
}

static struct ca_hfile *
ca_hfile_new(struct ca_db *db, char *path, int create,
	     int (*rmethod)(), int (*wmethod)())
{
	struct ca_hfile *hf;
	
	hf = malloc(sizeof(struct ca_hfile));
	if (hf == NULL) {
		perror("malloc");
		return(NULL);
	}
	hf->db = db;
	if (!is_file(path)) {
		if (create) {
			if (creat(path, 0666) < 0) {
				perror("creat");
				free(hf);
				return(NULL);
			}
		} else {
			free(hf);
			return(NULL);
		}
	}
	hf->fpath = strdup(path);
	hf->r = rmethod;
	hf->w = wmethod;
	hf->next = db->hfiles;
	db->hfiles = hf;

	return hf;
}

static int
ca_hfile_read(struct ca_hfile *h)
{
	if (h->r)
		return h->r(h);
	return -1;
}

static int
ca_hfile_write(struct ca_hfile *h)
{
	if (h->w)
		return h->w(h);
	return -1;
}

static void
ca_hfile_free(struct ca_hfile *hf)
{
	free(hf->fpath);
	free(hf);
}

static struct ca_user *
ca_user_new(char *name, char *pass,
	    char *sysuser, char *email)
{
	struct ca_user *user;

	if (name == NULL)
		return(NULL);

	user = malloc(sizeof(struct ca_user));
	if (user == NULL) {
		perror("malloc");
		return(NULL);
	}
	user->name = strdup(name);
	user->pass = pass ? strdup(pass) : NULL;
	user->sysuser = sysuser ? strdup(sysuser) : NULL;
	user->email = email ? strdup(email) : NULL;
	user->next = NULL;

	return user;
}

static void 
ca_user_free(struct ca_user *user)
{
	free(user->name);
	if (user->pass) free(user->pass);
	if (user->sysuser)free(user->sysuser);
	if (user->email)free(user->email);
	free(user);
}


static struct ca_db *repositories[MAX_REPOS];

static struct ca_db *
get_db(int id)
{
	if (id < 0 || id > MAX_REPOS)
		return NULL;
	return repositories[id];
}

static int
get_id(struct ca_db *db)
{
	int i;

	for (i = 0; i < MAX_REPOS; i++) {
		if (repositories[i] == db)
			return i; 
	}
	return -1;
}

static int
is_local_repository(const char *repos)
{
	if (repos[0] != ':')
		return 1;
	if (!strncmp(repos + 1, "local", 3))
		return 1;
	return 0;
}

/*
 * Public API
 */

/*
 * cvsadmin_open
 *
 *  Open a CVS repository for administration.  We return an identifier
 * to be used in subsequent cvsadmin_* calls.  If the <usecvs> flag is
 * set, we will use cvs(1) to administer the file, otherwise we will
 * directly access the file.  If <usecvs> is set, then <repos> must be
 * a local repository
 */
int
cvsadmin_open(char *repos, bool usecvs)
{
	int i;
	struct ca_db *db;

	/* XXX - Having the '-c' option of cvsadmin(1) mentionned in the
	 * backend is Not Good.  Maybe we should set a local error variable */
	if (!usecvs && !is_local_repository(repos)) {
		fprintf(stderr, "Not using cvs(1) and non-local repository `%s', can't continue, see the -c switch", repos);
		return (-1);
	}

	/* Find an unused repository slot */
	for(i = 0; i < MAX_REPOS; i++) {
		if (repositories[i] == NULL)
			break;
	}
	if (i == MAX_REPOS) {
		fprintf(stderr, "Too many openned repositories\n");
		return -1;
	}

	if ((db = ca_db_new(repos, usecvs)) == NULL)
		return -1;

	repositories[i] = db;

	/* Try to open a password file. */
	if (!ca_db_try_hfile(db, "/etc/cvs.passwd",
			     0, etc_cvs_passwd_read, etc_cvs_passwd_write)) {
		if (!ca_db_try_hfile(db, "passwd", 
				     1, passwd_read, passwd_write)) {
			printf("Couldn't read any password file in %s, probably\n"
			       "fatal\n", repos);
			ca_db_free(db);
			repositories[i] = NULL;

			return -1;
		}
	}
	ca_db_try_hfile(db, "users", 0, users_read, users_write);

	db->changed = 0; /* the add-file thing has changed the repository */

	return i;
}

/*
 * cvsadmin_set_user_info
 *
 *  Modify some information about a user. <id> is the repository
 * identifier as returned by cvsadmin_open().  If <uname> doesn't
 * exist, it is created.
 *
 *  We only set "fields" that have non-NULL values.
 */
int
cvsadmin_set_user_info(int id, char *uname, char *pass,
		       char *sysuser, char *email)
{
	struct ca_user *u;
	struct ca_db *db;

	db = get_db(id);
	if (db == NULL)
		return -1;

	u = ca_db_find_user(db, uname);
	if (u == NULL) {
		u = ca_user_new(uname, NULL, NULL, NULL);
		if (u)
			ca_db_add_user(db, u); 
	}

#define T(x) do { \
		  if (x){ if (u->x){free(u->x);}	\
		  u->x = strdup(x); db->changed++; } \
	    } while (0)
	T(pass);
	T(sysuser);
	T(email);
#undef T

	return 0;
}

/*
 * cvsadmin_del_user
 *
 * Delete a user from this repository. <id> is the repository ID
 * returned by the cvsadmin_open() call
 */
int
cvsadmin_del_user(int id, char *name)
{
	struct ca_db *db;
	struct ca_user *u;

	db = get_db(id);
	if (db == NULL)
		return -1;

	u = ca_db_find_user(db, name);
	if (u == NULL)
		return -1;

	if (ca_db_del_user(db, u) < 0)
		return -1;

	ca_user_free(u);

	return 0;
}

/*
 * cvsadmin_rename_user
 *
 * Rename this user.
 */
int
cvsadmin_rename_user(int id, char *oname, char *nname)
{
	struct ca_user *u;
	struct ca_db *db;

	db = get_db(id);
	if (db == NULL)
		return -1;

	/* find out if the new name is already in use */
	u = ca_db_find_user(db, nname);
	if (u != NULL){
		fprintf(stderr, "%s: user already exists\n", nname);
		return -1;
	}

	/* Note that this code will not work as soon as we put
	 * user in an hash table!! */
	u = ca_db_find_user(db, oname);
	if (u == NULL) {
		fprintf(stderr, "%s: user doesn't exist\n", oname);
		return -1;
	}
	free(u->name);
	u->name = strdup(nname);

	db->changed = 1;

	return 0;
}

/*
 * cvsadmin_get_user_info
 *
 *   Get information from a user. We give out only the information
 * that is requested by a non-NULL pointer.
 */
int
cvsadmin_get_user_info(int id, char *uname, char **Opass, char **Osysuser,
			   char **Oemail)
{
	struct ca_user *u;
	struct ca_db *db;

	db = get_db(id);
	if (db == NULL)
		return -1;

	u = ca_db_find_user(db, uname);
	if (u == NULL)
		return -1;

	if (Opass != NULL)	/* Do we want pass? */
		*Opass = u->pass;
	if (Osysuser != NULL)	/* Do we want sysuser? */
		*Osysuser = u->sysuser;
	if (Oemail != NULL)	/* Do we want email? */
		*Oemail = u->email;

	return 0;
}

/*
 * cvsadmin_get_user_list
 *
 *  Get a NULL-terminated list of all the users in repository <id>.
 * 
 *  cvsadmin_free_str_list() should be used to free this functions'
 * return value.
 */
char **
cvsadmin_get_user_list(int id)
{
	struct ca_db *db;
	struct ca_user *p;
	char **list;
	int n;

	db = get_db(id);
	if (db == NULL)
		return NULL;

	/* Are there any users? */
	n = db->nusers;
	if (n == 0) {
		fprintf(stderr, "No users in database\n");
		return NULL;
	}

	list = (char **)xmalloc((n + 1) * sizeof(char *));

	n = 0;
	for (p = db->users; p != NULL; p = p->next)
		list[n++] = p->name;

	list[n] = NULL;

	return list;
}

/*
 * cvsadmin_free_str_list
 *
 * Free a string list returned by cvsadmin_get_user_list
 */
void
cvsadmin_free_str_list(char **list)
{
	int i;

	if (list != NULL) {
		for (i = 0; list[i] != NULL; i++)
			free(list[i]);
		free(list);
	}
}

/*
 * cvsadmin_sync
 *
 * Sync out repository <id> by writing out changed stuff, etc.
 */
void
cvsadmin_sync(int id)
{
	struct ca_db *db;
	struct ca_hfile *hf;

	db = get_db(id);
	if (db == NULL)
		return;

	if (!db->changed)
		return;

	for (hf = db->hfiles; hf != NULL; hf=hf->next)
		ca_hfile_write(hf);

	if (db->usecvs) {
		if (db->changed)
			if (call_cvs4(db->container, db->repository, "commit",
				      "-mcvsadmin_change") < 0)
				fprintf(stderr, "Can't commit new files\n");
	}
	db->changed=0;
}

/*
 * cvsadmin_close
 *
 * Close this database and free all internal data associated with it.
 *
 * This calls cvsadmin_sync
 */
int
cvsadmin_close(int id)
{
	struct ca_db *db;
	struct ca_hfile *hf, *hfn;

	db = get_db(id);
	if (db == NULL)
		return -1;

	cvsadmin_sync(id);
 
	for (hf = db->hfiles; hf != NULL; hf = hfn) {
		hfn = hf->next;
		ca_hfile_free(hf);
	}
	if (db->usecvs) {
		if (rm_rf(db->container) < 0)
			fprintf(stderr, "***Warning: Couldn't clean up temp"
				"directory\n");
	}

	
	ca_db_free(db);

	repositories[id]=0;
	
	return 0;
}

/*
 * cvsadmin_cleanup
 *
 * This is a function used to close all open repositories.
 */
void
cvsadmin_cleanup(void)
{
	int i;

	for (i = 0; i < MAX_REPOS; i++)
		cvsadmin_close(i);
}

