/*
 * util.c - Misc utility functions for cvsadmin
 *
 * Copyright (c) 2001, Vincent Labrecque <limitln@cooptel.qc.ca>
 * All rights reserved.
 * 
 *      Redistribution and use in source and binary forms,
 *      with or without modification, are permitted provided
 *      that the following conditions are met:
 * 
 *      Redistributions of source code must retain the
 *      above copyright notice, this list of conditions
 *      and the following disclaimer. 
 *      Redistributions in binary form must reproduce
 *      the above copyright notice, this list of
 *      conditions and the following disclaimer in the
 *      documentation and/or other materials provided
 *      with the distribution. 
 *      Neither the name of the <ORGANIZATION> nor the
 *      names of its contributors may be used to
 *      endorse or promote products derived from this
 *      software without specific prior written
 *      permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/wait.h>

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

/*	
 * Calls cvs with arguments
 *	"cvs -d repository command param"
 *
 * (0 means OK, <0 means error)
 */
int call_cvs3(char *repository, char *command, char *param)
{
	char *argv[] = { "/usr/bin/cvs", "-d", NULL, NULL, NULL, NULL };
	pid_t pid;
	int ret, i;
	
	if (command == NULL)
		command = "";
	if (param == NULL)
		param = "";
	for (i = 0; argv[i] != NULL; i++)
		;
	argv[i++] = repository;
	argv[i++] = command;
	argv[i++] = param;

	pid = vfork();
	if (pid < 0) {
		perror("vfork");
		exit(1);
	} else if (pid == 0) {
		execv(argv[0], argv);
		perror("execv");
		_exit(1);
	} else {
		int status;
		while (waitpid(pid,&status,0) > 0)
			;
		ret = 0;
		if (WIFEXITED(status)) {
			if (WEXITSTATUS(status) == 1)
				ret = -1;
		} else if (WIFSIGNALED(status))
			ret = -1;
	}
	return ret;
}

int call_cvs4(dir, repository, command, param)
	char *dir, *repository, *command, *param;
{
	if (chdir(dir) < 0)
		return -1;
	return call_cvs3(repository, command, param);
}

/*
 * Simulates a rm -rf on path
 *  
 * returns -1 if there was any error.
 *
 *  Note: We don't back if there is an error, if it returns -1 you may have
 *        a only few files deleted, etc.
 *
 * XXX: This doesn't look safe to me... but well, there is no fchdir,
 *      funlink or anything similar.  If someone can do a better implementation
 *      of this, send it to me, please!
 */
int rm_rf(char *path)
{
	struct stat st;
	DIR *dir;
	struct dirent *fi;
	char buf[FILENAME_MAX], *p;
	int avail;
	
	if (lstat(path, &st) < 0)
		return -1;
	
	if (S_ISDIR(st.st_mode)) {

		if ((dir = opendir(path)) == NULL) {
			perror("opendir");
			return -1;
		}
		p = buf + snprintf(buf, sizeof(buf), "%s/", path);
		avail = sizeof(buf) - (p - buf);
		
		while ((fi = readdir(dir)) != NULL) {
			if (!strcmp(fi->d_name, ".") ||
			    !strcmp(fi->d_name, ".."))
				continue;

			strncpy(p, fi->d_name, avail);

			if (rm_rf(buf) < 0)
				return -1;
		}
		if (closedir(dir) < 0)
			perror("closedir");

		if (rmdir(path) < 0)
			return -1;
	}
	else {
		if (unlink(path) < 0) {
			perror("unlink");
			return -1;
		}
	}

	return 0;
}


/*
 * Remove trailing whitespaces from str, returns a pointer to the first
 * non-blank character
 */
char *clean_string(char *str)
{
	char *p;

	/* find the real end */
	p = str + strlen(str) -1;
	while (p >= str) {
		if(!isspace(*p))
			break;
		*p-- = '\0';		/* chop chop! */
	}
	if (p == str)
		return NULL; /*only spaces*/

	/* find the real start */
	for (p = str; *p && isspace(*p); p++)
		;

	return p;
}


/* code by der Mouse */
/* salt must be 3 bytes long */
void generate_salt(char *salt)
{
 static const char alphabet[64] = "abcdefghijklmnopqrstuvwxyz"
				  "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
				  "0123456789./";
 static int didinit = 0;

 if (! didinit)
  { time_t now;
    time(&now);
    srand(now);
    didinit=1;
  }
 salt[0] = alphabet[rand()&63];
 salt[1] = alphabet[rand()&63];
 salt[2] = '\0';
}

int is_file(char *path)
{
 struct stat st;

 if (stat(path,&st)<0)return 0;
 if (!S_ISREG(st.st_mode)) return 0;

 return 1;
}

#ifdef HAVE_MKDTEMP
# define my_mkdtemp mkdtemp
#else
/*
 * This one is really crappy.
 * It's also less secure.  But if you care about security,
 * you should be running OpenBSD anyways, which has mkdtemp :)
 *
 * TODO: rewrite this thing so it makes some sense!
 */
char *my_mkdtemp(temp)
	char *temp;
{
	char *sp, *p;
	struct stat st;

	sp = temp + strlen(temp) - 1;
	while (sp >= temp && *sp == 'X')
		sp--;
	srand(time(NULL)^getpid());
	for (;;) {
		/* eurk */
		for(p = sp; *p != 0; p++)
			*p = rand()%('z'-'a')+'a';
		if(stat(temp, &st)<0 && errno==ENOENT) break;
	}
	if(mkdir(temp, 0700)<0) return NULL;
	return temp;
}
#endif /* MKDTEMP */

char *get_temp_dir()
{
	char temp[] = "/tmp/.ca_tmpXXXXXX";
	if (!my_mkdtemp(temp)) return NULL;
	return strdup(temp);
}

char *extract_path(char *repos)
{
	char *p;

	if(*repos == ':') {
		p = index(repos+1, ':');
		if(!p) return NULL;
		p++;
	} else p = repos;

	return p;
}

int my_rename(char *source, char *dest)
{
	char buf[BUFSIZ];
	FILE *i, *o;
	int n;

	if((n = rename(source, dest)) == -1 && errno == EXDEV) {
		i = fopen(source, "r");
		if(!i) {
			perror(source);
			return -1;
		}
		o = fopen(dest, "w");
		if(!o) {
			perror(dest);
			return -1;
		}
 		while ((n = fread(buf, sizeof(buf[0]), sizeof(buf), i)) > 0) {
			if (fwrite(buf, sizeof(buf[0]), n, o) < n) {
				perror("fwrite");
			}
		}
		if(fclose(i) < 0)
			perror("fclose input");
		if(fclose(o) < 0) {
			n = -1;
			perror("fclose output");
		} else {
			if(unlink(source) < 0)
				perror("unlink");
			n = 0;
		}
	}

	return n;
}
